import os
from collections import OrderedDict

from . import LayerSparsity


class Hoplite:
    """Hoplite Sparsity Analyzer"""

    def __init__(
        self, model, preprocess, layers=[], zero_sensitivity=0, total_max=None
    ):
        """Hoplite class constructor

        model - Keras Model to be analyzed
        preprocess - a function that takes in a filename and returns input run the model on
        layers - a list of the names of the layers to analyze outputs of
        zero_sensitivity (optional) - how sensitive to be when checking for zeroes
        total_max (optional) - total number of inputs Hoplite will accept, it will skip the rest"""

        self.model = model
        self.preprocess = preprocess
        self.layers = layers
        self.zero_sensitivity = zero_sensitivity
        self.total_max = total_max
        self.counter = 0  # number of times analysis has run

        self.sparsities = OrderedDict.fromkeys(self.layers, [])

    def equals_zero(self, number):
        """Checks if a given number is considered zero"""
        return abs(number) < self.zero_sensitivity

    def exceeded_max(self):
        """Checks if exceeded max yet"""
        return self.total_max is not None and self.counter > self.total_max

    def analyze_file(self, filename):
        """Analyze a given file"""
        if self.exceeded_max():
            return

        if self.preprocess is not None:
            input = self.preprocess(filename)
        else:
            with open(filename, "r") as file:
                input = file.read()

        self.analyze_raw(input)

    def analyze_dir(self, dirname):
        """Analyze a directory of files"""
        if self.exceeded_max():
            return

        for (dirpath, dirnames, filenames) in os.walk(dir_name):
            for filename in filenames:
                self.analyze_file(dirname + "/" + filename)

    def analyze_raw(self, input):
        """Analyze raw input"""
        if self.exceeded_max():
            return

        for layer in self.layers:
            layer_s = LayerSparsity(layer, self.model.get_layer(layer).output_shape[1:])

            layer_s.set_sparsities(self.model, input, equals_zero=self.equals_zero)

            self.sparsities[layer].append(layer_s)

    def output(self, filename):
        for layer in self.sparsities:
            LayerSparsity.average(self.sparsities[layer]).output(filename)



import numpy as np


class Spartan:
    """Sparsity Analysis Helper"""

    ROW_AXIS = 2
    COL_AXIS = 0
    CHAN_AXIS = 1

    @staticmethod
    def chunk_array(lst, n):
        """Breaks up arrays into n sized chunks, last chunk may be less than n
        if total length is not a multiple of n"""
        for i in range(0, len(lst), n):
            yield lst[i : i + n]

    @staticmethod
    def compute_average_sparsity(output, equals_zero=lambda x: x == 0):
        """Computes average sparsity of whole output"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        count = 0
        for chan in output:
            for col in chan:
                for number in col:
                    if equals_zero(number):
                        count += 1
        return float(count) / output.size

    @staticmethod
    def consec_1d(arr, hist, equals_zero=lambda x: x == 0):
        """calculates consecutive vectors of zeroes"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        all_nonzeroes = True
        count = 0
        for a in range(len(arr)):
            if equals_zero(arr[a]):
                all_nonzeroes = False
                count += 1
            else:
                if count != 0:
                    hist[count] += 1
                    count = 0
        if count != 0:
            hist[count] += 1
        if all_nonzeroes:
            hist[0] += 1

    @staticmethod
    def consec_chan(output, equals_zero=lambda x: x == 0):
        """Calculate consecutive zeros histogram in a channel"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        chan_hist = [0] * (len(output[0][0]) + 1)
        np.apply_along_axis(
            Spartan.consec_1d, Spartan.CHAN_AXIS, output, chan_hist, equals_zero
        )
        return chan_hist

    @staticmethod
    def consec_row(output, equals_zero=lambda x: x == 0):
        """Calculate consecutive zeros histogram in a row"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        row_hist = [0] * (len(output[0]) + 1)
        np.apply_along_axis(
            Spartan.consec_1d, Spartan.ROW_AXIS, output, row_hist, equals_zero
        )
        return row_hist

    @staticmethod
    def consec_col(output, equals_zero=lambda x: x == 0):
        """Calculate consecutive zeros histogram in a column"""
        global COL
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        col_hist = [0] * (len(output) + 1)
        np.apply_along_axis(
            Spartan.consec_1d, Spartan.COL_AXIS, output, col_hist, equals_zero
        )
        return col_hist

    @staticmethod
    def vec_1d(arr, vec_size, hist, equals_zero=lambda x: x == 0):
        """Calculates sparsity given a vector size"""
        if len(arr) < vec_size:
            return

        if equals_zero is None:
            equals_zero = lambda x: x == 0

        chunks = Spartan.chunk_array(arr, vec_size)
        for chunk in chunks:
            zeroes = 0
            for num in chunk:
                if equals_zero(num):
                    zeroes += 1
            hist[zeroes] += 1

    @staticmethod
    def vec_3d_chan(output, vec_size, equals_zero=lambda x: x == 0):
        """Calculates sparsity of a channel given a vector size"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        vec_chan_hist = [0] * (vec_size + 1)
        np.apply_along_axis(
            Spartan.vec_1d, Spartan.CHAN_AXIS, output, vec_size, vec_chan_hist
        )
        return vec_chan_hist

    @staticmethod
    def vec_3d_row(output, vec_size, equals_zero=lambda x: x == 0):
        """Calculates sparsity of a row given a vector size"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        vec_row_hist = [0] * (vec_size + 1)
        np.apply_along_axis(
            Spartan.vec_1d, Spartan.ROW_AXIS, output, vec_size, vec_row_hist
        )
        return vec_row_hist

    @staticmethod
    def vec_3d_col(output, vec_size, equals_zero=lambda x: x == 0):
        """Calculates sparsity of a column given a vector size"""
        if equals_zero is None:
            equals_zero = lambda x: x == 0

        vec_col_hist = [0] * (vec_size + 1)
        np.apply_along_axis(
            Spartan.vec_1d, Spartan.COL_AXIS, output, vec_size, vec_col_hist
        )
        return vec_col_hist



import csv
import sys
from collections import OrderedDict

import numpy as np
from tensorflow.keras.models import Model

from . import Spartan


class LayerSparsity:
    def __init__(self, name, dimensions, vector_sizes=[2, 4, 8, 16, 32]):
        self.name = name
        self.dimensions = dimensions
        self.avg_sparsity = 0
        self.vector_sizes = vector_sizes

        # histograms is an OrderedDict for ease of output
        self.histograms = OrderedDict.fromkeys(
            ["row_hist", "col_hist", "chan_hist",]
            + [
                text.format(vec)
                for vec in vector_sizes
                for text in ["vec{}_row_hist", "vec{}_col_hist", "vec{}_chan_hist"]
            ],
            [],  # empty array is default value for all histograms
        )

    @staticmethod
    def average(layer_list):
        """Returns the average LayerSparsity from multiple LayerSparities"""

        if len(layer_list) < 1:
            print("error: averaging zero layers!")
            return

        if len(layer_list) == 1:
            return layer_list[0]

        if (
            layer_list[0].name != layer_list[1].name
            or layer_list[0].dimensions != layer_list[1].dimensions
        ):
            print("error: averaging different layers!")
            return

        out = LayerSparsity(self.name, self.dimensions, layer_list[0].vector_sizes)

        for key in layer_list[0].histograms:
            out.histograms[key] = np.mean(
                np.array([layer.histograms[key] for layer in layer_list]), axis=0
            ).tolist()

        return out

    def set_sparsities(self, model, input, equals_zero=lambda x: x == 0):
        layer_model = Model(
            inputs=model.inputs, outputs=model.get_layer(self.name).output
        )

        output = layer_model.predict(input)[0]
        # TODO finish setting sparsities
        self.avg_sparsity = Spartan.compute_average_sparsity(output, equals_zero)

        self.histograms["row_hist"] = Spartan.consec_row(output, equals_zero)
        self.histograms["col_hist"] = Spartan.consec_col(output, equals_zero)
        self.histograms["chan_hist"] = Spartan.consec_chan(output, equals_zero)

        for size in self.vector_sizes:
            self.histograms["vec{}_row_hist".format(size)] = Spartan.vec_3d_row(
                output, size, equals_zero
            )
            self.histograms["vec{}_col_hist".format(size)] = Spartan.vec_3d_col(
                output, size, equals_zero
            )
            self.histograms["vec{}_chan_hist".format(size)] = Spartan.vec_3d_chan(
                output, size, equals_zero
            )

    def output(self, filename=None):
        """Output contents of LayerSparsity object to a given filename, default: stdout"""
        if filename is None:
            file = sys.stdout
        else:
            file = open(filename, "a+")
        writer = csv.writer(file, delimiter=",")

        writer.writerow(["layer=", self.name])
        writer.writerow(["dimensions=", self.dimensions])
        writer.writerow(["average=", self.avg_sparsity])

        for key in self.histograms:
            writer.writerow(["{}=".format(key), self.histograms[key]])

        file.close()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hoplite2']

package_data = \
{'': ['*']}

install_requires = \
['Pillow >= 7.2.0', 'tensorflow >= 2.2.0']

setup(name='hoplite2',
      version='0.2.0',
      description='A Sparsity Analysis Tool and Library',
      author='Pranoy Dutta',
      author_email='pranoydll@gmail.com',
      url='https://github.com/prydt/hoplite2',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )
